@@ -4,7 +4,7 @@ JSONLD - A toolkit for interacting with JSON-LD data.
 
 =head1 VERSION
 
-This document describes JSONLD version 0.000_03
+This document describes JSONLD version 0.000_04
 
 =head1 SYNOPSIS
 
@@ -34,7 +34,7 @@ linked data.
 package JSONLD {
 	use v5.18;
 	use autodie;
-	our $VERSION	= '0.000_03';
+	our $VERSION	= '0.000_04';
 	use utf8;
 	use Moo;
 	use LWP;
@@ -88,6 +88,40 @@ Returns the JSON-LD expansion of C<< $data >>.
 		return $self->_expand($ctx, undef, $d);
 	}
 	
+	sub _expand {
+		my $self	= shift;
+		my $ctx		= shift // {};
+		my $prop	= shift;
+		my $d		= shift;
+		my $expanded_output	= $self->_5_1_2_expansion($ctx, $prop, $d, @_);
+		if (ref($expanded_output) eq 'HASH') {
+			my @keys	= keys %$expanded_output;
+			if (scalar(@keys) == 1 and $keys[0] eq '@graph') {
+				println "_expand 7.1" if $debug;
+				$expanded_output	= $expanded_output->{'@graph'};
+			}
+		}
+		
+		unless (defined($expanded_output)) {
+			println "_expand 7.2" if $debug;
+			$expanded_output	= [];
+		}
+		
+		if (ref($expanded_output) ne 'ARRAY') {
+			println "_expand 7.3" if $debug;
+			$expanded_output	= [$expanded_output];
+		}
+		
+		return $expanded_output;
+	}
+	
+	sub _is_scalar {
+		my $v	= shift;
+		return 1 unless (ref($v));
+		return 1 if JSON::is_bool($v);
+		return 0;
+	}
+	
 	sub _is_string {
 		my $v	= shift;
 		return 0 unless defined($v);
@@ -116,15 +150,6 @@ Returns the JSON-LD expansion of C<< $data >>.
 		return $is_num;
 	}
 	
-	sub _expand {
-		my $self	= shift;
-		my $ctx		= shift // {};
-		my $prop	= shift;
-		my $d		= shift;
-		my $v	= $self->_5_1_2_expansion($ctx, $prop, $d, @_);
-		return $v;
-	}
-	
 	sub _is_abs_iri {
 		my $self	= shift;
 		my $value	= shift;
@@ -143,9 +168,6 @@ Returns the JSON-LD expansion of C<< $data >>.
 		my $i		= eval { IRI->new(value => $value) };
 		my $err		= $@;
 		my $is_iri	= not($@);
-		if ($@) {
-			Carp::confess $@;
-		}
 		return $is_iri;
 	}
 	
@@ -479,7 +501,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 				if (not defined($value)) {
 					println "5.10.3" if $debug;
 					delete $result->{'@direction'};
-				} elsif (not ref($value)) {
+				} elsif (_is_string($value)) {
 					println "5.10.4 \@direction = $value" if $debug;
 					if ($value ne 'ltr' and $value ne 'rtl') {
 						die 'invalid base direction';
@@ -696,7 +718,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 		println "15" if $debug;
 		$definition->{'reverse'}	= 0; # 15
 		
-		if (exists $value->{'@id'} and $value->{'@id'} ne $term) {
+		if (exists $value->{'@id'} and (not(defined($value->{'@id'})) or $value->{'@id'} ne $term)) {
 			# 16
 			println "16" if $debug;
 			if (not defined($value->{'@id'})) {
@@ -990,7 +1012,6 @@ Returns the JSON-LD expansion of C<< $data >>.
 		my $property_scoped_ctx;
 		my $property_scoped_ctx_defined	= 0;
 		my $tdef = $self->_ctx_term_defn($activeCtx, $activeProp);
-		warn Dumper($tdef);
 		if ($tdef and exists $tdef->{'@context'}) {
 			$property_scoped_ctx_defined	= 1;
 			my $lctx = $tdef->{'@context'};
@@ -998,7 +1019,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 			println "3 property-scoped context for property $activeProp: " . Data::Dumper->Dump([$property_scoped_ctx], [qw(context)]) if $debug;
 		}
 		
-		if (not(ref($element))) {
+		if (_is_scalar($element)) {
 			# 4
 			println "4" if $debug;
 			if (not(defined($activeProp)) or $activeProp eq '@graph') {
@@ -1077,7 +1098,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 			}
 			local($Data::Dumper::Indent)	= 1;
 			println "before 8: " . Data::Dumper->Dump([$activeCtx, $property_scoped_ctx], [qw'activeCtx property_scoped_ctx']) if $debug;
-			$activeCtx = $self->_4_1_2_ctx_processing($activeCtx, $property_scoped_ctx, %args); # 8
+			$activeCtx = $self->_4_1_2_ctx_processing($activeCtx, $property_scoped_ctx, %args, override_protected => 1); # 8
 			println "after 8: " . Data::Dumper->Dump([$activeCtx], ['activeCtx']) if $debug;
 		}
 		
@@ -1108,12 +1129,16 @@ Returns the JSON-LD expansion of C<< $data >>.
 				$value	= [$value]; # 11.1
 			}
 			
+# 			my %tdefs	= map { $_ => $self->_ctx_term_defn($activeCtx, $_) } grep { _is_string($_) }@$value; # https://github.com/w3c/json-ld-api/issues/304
+			my %tdefs	= map { $_ => $self->_ctx_term_defn($type_scoped_ctx, $_) } grep { _is_string($_) }@$value; # https://github.com/w3c/json-ld-api/issues/304
 			foreach my $term (sort @$value) {
 				println "11.2 attempting with [$term]" if $debug;
-				if (not(ref($term))) {
-					my $tdef	= $self->_ctx_term_defn($activeCtx, $term);
-					if (my $c = $tdef->{'@context'}) {
+				if (_is_string($term)) {
+					my $tdef	= $tdefs{$term};
+# 					my $tdef	= $self->_ctx_term_defn($activeCtx, $term);
+					if (exists $tdef->{'@context'}) {
 						println "11.2" if $debug;
+						my $c = $tdef->{'@context'};
 						$activeCtx	= $self->_4_1_2_ctx_processing($activeCtx, $c, propagate => 0);
 						local($Data::Dumper::Indent)	= 1;
 						println "11.2 " . Data::Dumper->Dump([$activeCtx], ['activeCtx']) if $debug;
@@ -1227,25 +1252,8 @@ Returns the JSON-LD expansion of C<< $data >>.
 			println "19 resulting in " . Data::Dumper->Dump([$result], ['*result']) if $debug;
 		}
 		
-		if (ref($result) eq 'HASH') {
-			if (scalar(@keys) == 1 and $keys[0] eq '@graph') {
-				println "20" if $debug;
-				$result	= $result->{'@graph'};
-			}
-		}
-		
-		unless (defined($result)) {
-			println "21" if $debug;
-			$result	= [];
-		}
-		
-		if (ref($result) ne 'ARRAY') {
-			println "22" if $debug;
-			$result	= [$result];
-		}
-		
 		local($Data::Dumper::Indent)	= 1;
-		println "23 returning from _5_1_2_expansion with final value " . Data::Dumper->Dump([$result], ['*result']) if $debug;
+		println "20 returning from _5_1_2_expansion with final value " . Data::Dumper->Dump([$result], ['*result']) if $debug;
 		return $result; # 19
 	}
 	
@@ -1318,7 +1326,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 					println "13.4.4" if $debug;
 					my $is_string = _is_string($value);
 					my $is_array	= ref($value) eq 'ARRAY';
-					my $is_array_of_strings	= ($is_array and all { not(ref($_)) } @$value);
+					my $is_array_of_strings	= ($is_array and all { _is_string($_) } @$value);
 					if (not($is_string) and not($is_array_of_strings)) {
 						println "13.4.4.1 invalid" if $debug;
 						die 'invalid type value';
@@ -1355,8 +1363,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 				if ($expandedProperty eq '@graph') {
 					println "13.4.5" if $debug;
 					my $v	= $self->_5_1_2_expansion($activeCtx, '@graph', $value, frameExpansion => $frameExpansion, ordered => $ordered);
-					# TODO: ensure that expanded value is an array of one or more maps
-					$expandedValue	= $v;
+					$expandedValue	= (ref($v) eq 'ARRAY') ? $v : [$v];
 					println "13.4.5 resulting in " . Data::Dumper->Dump([$expandedValue], ['*expandedValue']) if $debug;
 				}
 
@@ -1393,7 +1400,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 						if ($self->processing_mode eq 'json-ld-1.0') {
 							die 'invalid value object value';
 						}
-					} elsif (not (not(ref($value)) or not defined($value))) { # "if value is not a scalar or null, an invalid value object value error has been detected"
+					} elsif (not (_is_scalar($value) or not defined($value))) { # "if value is not a scalar or null, an invalid value object value error has been detected"
 						println "13.4.7.2 " .  Data::Dumper->Dump([$value], ['*value']) if $debug; # NOTE: the language here is ambiguous: "if value is not a scalar or null"
 						die 'invalid value object value';
 					} else {
@@ -1489,59 +1496,56 @@ Returns the JSON-LD expansion of C<< $data >>.
 					$expandedValue	= $self->_5_1_2_expansion($activeCtx, '@reverse', $value, frameExpansion => $frameExpansion, ordered => $ordered); # 13.4.13.1
 					println "13.4.13.2 " . Data::Dumper->Dump([$expandedValue], ['*expandedValue']) if $debug;
 					
-					foreach my $ev (@{ $expandedValue }) { # https://github.com/w3c/json-ld-api/issues/294
-						my $expandedValue	= $ev;
-						if (ref($expandedValue) eq 'HASH' and exists $expandedValue->{'@reverse'}) { # NOTE: spec text does not assert that expandedValue is a map
-							println "13.4.13.3" if $debug;
-							foreach my $property (keys %{ $expandedValue->{'@reverse'} }) {
-								my $__indent	= indent();
-								println "13.4.13.3 [$property]" if $debug;
-								my $item	= $expandedValue->{'@reverse'}{$property};
-								if (not exists $result->{$property}) {
-									println "13.4.13.3.1" if $debug;
-									$result->{$property}	= [];
-								}
-							
-								println "13.4.13.3.2" if $debug;
-								push(@{ $result->{$property} }, @$item);
+					if (ref($expandedValue) eq 'HASH' and exists $expandedValue->{'@reverse'}) { # NOTE: spec text does not assert that expandedValue is a map
+						println "13.4.13.3" if $debug;
+						foreach my $property (keys %{ $expandedValue->{'@reverse'} }) {
+							my $__indent	= indent();
+							println "13.4.13.3 [$property]" if $debug;
+							my $item	= $expandedValue->{'@reverse'}{$property};
+							if (not exists $result->{$property}) {
+								println "13.4.13.3.1" if $debug;
+								$result->{$property}	= [];
 							}
+						
+							println "13.4.13.3.2" if $debug;
+							push(@{ $result->{$property} }, @$item);
 						}
+					}
+				
+					if (ref($expandedValue) eq 'HASH') { # NOTE: spec text does not assert that expandedValue is a map
+						my @keys	= grep { $_ ne '@reverse' } keys %$expandedValue;
+						if (scalar(@keys)) {
+							println "13.4.13.4" if $debug;
 					
-						if (ref($expandedValue) eq 'HASH') { # NOTE: spec text does not assert that expandedValue is a map
-							my @keys	= grep { $_ ne '@reverse' } keys %$expandedValue;
-							if (scalar(@keys)) {
-								println "13.4.13.4" if $debug;
+							if (not exists $result->{'@reverse'}) {
+								println "13.4.13.4.1" if $debug;
+								$result->{'@reverse'}	= {};
+							}
+					
+							println "13.4.13.4.2" if $debug;
+							my $reverse_map	= $result->{'@reverse'};
+					
+							println "13.4.13.4.3" if $debug;
+							foreach my $property (grep { $_ ne '@reverse' } keys %{ $expandedValue }) {
+								my $__indent	= indent();
+								println "13.4.13.4.3 [$property]" if $debug;
+								my $items	= $expandedValue->{$property};
 						
-								if (not exists $result->{'@reverse'}) {
-									println "13.4.13.4.1" if $debug;
-									$result->{'@reverse'}	= {};
-								}
-						
-								println "13.4.13.4.2" if $debug;
-								my $reverse_map	= $result->{'@reverse'};
-						
-								println "13.4.13.4.3" if $debug;
-								foreach my $property (grep { $_ ne '@reverse' } keys %{ $expandedValue }) {
+								println "13.4.13.4.3.1" if $debug;
+								foreach my $item (@$items) {
 									my $__indent	= indent();
-									println "13.4.13.4.3 [$property]" if $debug;
-									my $items	= $expandedValue->{$property};
-							
-									println "13.4.13.4.3.1" if $debug;
-									foreach my $item (@$items) {
-										my $__indent	= indent();
-										if ($self->_is_value_object($item) or $self->_is_list_object($item)) {
-											println "13.4.13.4.3.1.1" if $debug;
-											die 'invalid reverse property value';
-										}
-								
-										if (not exists $reverse_map->{$property}) {
-											println "13.4.13.4.3.1.2" if $debug;
-											$reverse_map->{$property}	= [];
-										}
-								
-										println "13.4.13.4.3.1.3" if $debug;
-										push(@{ $reverse_map->{$property} }, $item);
+									if ($self->_is_value_object($item) or $self->_is_list_object($item)) {
+										println "13.4.13.4.3.1.1" if $debug;
+										die 'invalid reverse property value';
 									}
+							
+									if (not exists $reverse_map->{$property}) {
+										println "13.4.13.4.3.1.2" if $debug;
+										$reverse_map->{$property}	= [];
+									}
+							
+									println "13.4.13.4.3.1.3" if $debug;
+									push(@{ $reverse_map->{$property} }, $item);
 								}
 							}
 						}
@@ -1565,7 +1569,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 					}
 				}
 				
-				println(Data::Dumper->Dump([$expandedValue, $expandedProperty, $input_type], [qw'expandedValue expandedProperty input_type'])) if $debug;
+				println "before 13.4.16 " . Data::Dumper->Dump([$expandedValue, $expandedProperty, $input_type], [qw'expandedValue expandedProperty input_type']) if $debug;
 				unless (not(defined($expandedValue)) and $expandedProperty eq '@value' and $input_type ne '@json') {
 					println "13.4.16 setting " . Data::Dumper->Dump([$expandedValue], ['*expandedValue']) if $debug;
 # 						println "$expandedProperty expanded value is " . Dumper($expandedValue) if $debug;
@@ -1690,7 +1694,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 					}
 					
 					println "13.8.3.6" if $debug;
-					$index_value	= $self->_expand($map_context, $key, $index_value, frameExpansion => $frameExpansion, ordered => $ordered);
+					$index_value	= $self->_5_1_2_expansion($map_context, $key, $index_value, frameExpansion => $frameExpansion, ordered => $ordered);
 					println(Data::Dumper->Dump([$index_value], ['*index_value'])) if $debug;
 					
 					println "13.8.3.7" if $debug;
@@ -1698,7 +1702,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 						my $__indent	= indent();
 						println '-----------------------------------------------------------------' if $debug;
 						println "13.8.3.7 [$item]" if $debug;
-						if ($self->_cm_contains($container_mapping, '@graph')) {
+						if ($self->_cm_contains($container_mapping, '@graph') and not $self->_is_graph_object($item)) {
 							println(Data::Dumper->Dump([$container_mapping], ['*container_mapping'])) if $debug;
 							println "13.8.3.7.1" if $debug;
 							$item	= {'@graph' => (ref($item) eq 'ARRAY') ? $item : [$item]};
@@ -1949,7 +1953,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 		}
 
 		if (my $tdef = $self->_ctx_term_defn($activeCtx, $value)) {
-			my $i	= $tdef->{'iri_mapping'};
+			my $i	= $tdef->{'iri_mapping'} // '';
 			if ($keywords{$i}) {
 				println "4 returning from _5_2_2_iri_expansion with a keyword: $i" if $debug;
 				return $i; # 4
@@ -2041,7 +2045,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 		if (exists($tdef->{'type_mapping'}) and $tm ne '@id' and $tm ne '@vocab' and $tm ne '@none') {
 			println "4" if $debug;
 			$result->{'@type'}	= $tm; # 4
-		} elsif (_is_string($value)) { # not(ref($value))) {
+		} elsif (_is_string($value)) {
 			println "5" if $debug;
 			println "5.1" if $debug;
 			my $language	= (exists $tdef->{'language_mapping'}) ? $tdef->{'language_mapping'} : $activeCtx->{'@language'}; # 5.1
@@ -2095,6 +2099,8 @@ RDF-related methods:
 
 =back
 
+See L<AtteanX::Parser::JSONLD> for an API that provides this functionality.
+
 =cut
 
 	sub to_rdf {
@@ -2496,7 +2502,9 @@ RDF-related methods:
 									$o,
 									$graph_iri
 								);
-								$self->add_quad($q, $dataset);
+								if ($q) {
+									$self->add_quad($q, $dataset);
+								}
 							}
 							println "1.3.2.5.4" if $debug;
 							foreach my $t (@$list_triples) {
@@ -2561,7 +2569,13 @@ RDF-related methods:
 			$datatype	= 'http://www.w3.org/1999/02/22-rdf-syntax-ns#JSON';
 		}
 		
-		println "9 TODO bool? " . Data::Dumper->Dump([$value], ['value']) if $debug;
+		if (JSON::is_bool($value)) {
+			println "9" if $debug;
+			$value	= $value ? 'true' : 'false';
+			unless ($datatype) {
+				$datatype	= 'http://www.w3.org/2001/XMLSchema#boolean';
+			}
+		}
 		
 		if (_is_numeric($value) and (not(_is_integer($value)) or (defined($datatype) and $datatype eq 'http://www.w3.org/2001/XMLSchema#double'))) {
 			println "10" if $debug;
@@ -2570,11 +2584,15 @@ RDF-related methods:
 			$value	=~ s/(\d)0E/$1E/;
 			$value	=~ s/E[+]0+/E0/;
 			$value	=~ s/E0+([1-9])/E$1/;
-			$datatype	= 'http://www.w3.org/2001/XMLSchema#double';
+			unless ($datatype) {
+				$datatype	= 'http://www.w3.org/2001/XMLSchema#double';
+			}
 		} elsif (_is_integer($value) or (_is_numeric($value) and (defined($datatype) and $datatype eq 'http://www.w3.org/2001/XMLSchema#integer'))) {
 			println "11" if $debug;
 			$value	= sprintf('%d', $value);
-			$datatype	= 'http://www.w3.org/2001/XMLSchema#integer';
+			unless ($datatype) {
+				$datatype	= 'http://www.w3.org/2001/XMLSchema#integer';
+			}
 		}
 
 		if (not defined($datatype)) {
@@ -2764,6 +2782,8 @@ at L<https://github.com/kasei/perl-jsonld/issues>.
 
 =item L<https://www.w3.org/TR/json-ld-api/>
 
+=item L<MooX::Role::JSON_LD>
+
 =back
 
 =head1 AUTHOR
